import argparse

from . import __version__
from .backend import Backend, DeleteOptions, RunOptions
from .engine import ArgparseEngine, Command


class kodmanEngine(ArgparseEngine):
    def __init__(self):
        if debug := self.get_env("KODMAN_DEBUG", bool):
            super().__init__(debug=debug)
        else:
            super().__init__()

        self.get_env("KODMAN_SERVICE_ACCOUNT", str)
        self._parser.add_argument(
            "-v",
            "--version",
            action="version",
            version=__version__,
        )
        self._ctx = Backend(self._log)


engine = kodmanEngine()


@engine.add_command
class Run(Command):
    def add(self, parser):
        parser_run = parser.add_parser("run", help="Run a command in a new container")
        parser_run.add_argument(
            "--entrypoint",
            type=str,
            help="Overwrite the default ENTRYPOINT of the image",
        )
        parser_run.add_argument(
            "--rm",
            help="Remove the container after exit",
            action="store_true",
        )
        parser_run.add_argument(
            "--volume",
            "-v",
            type=str,
            action="append",
            help="Bind mount a volume into the container",
        )
        parser_run.add_argument("image")
        parser_run.add_argument("command", nargs="?")
        parser_run.add_argument("args", nargs=argparse.REMAINDER, default=[])

    def do(self, args, ctx, env, log):
        ctx.connect()
        log.debug(f"Image: {args.image}")
        pod_name = ""
        k8s_command = []
        k8s_args = []
        if args.entrypoint:
            k8s_command = [args.entrypoint]
        if args.command:
            k8s_args.append(args.command)
        if args.args:
            k8s_args += args.args

        log.debug(f"Command: {k8s_command}")
        log.debug(f"Args: {k8s_args}")

        service_a = env["KODMAN_SERVICE_ACCOUNT"]
        options = RunOptions(
            image=args.image,
            command=k8s_command,
            args=k8s_args,
            volumes=args.volume,
            service_account=service_a if service_a else "",
        )

        pod_name = ctx.run(options)
        self.exit_code = ctx.return_code
        if args.rm:
            ctx.delete(DeleteOptions(pod_name))


@engine.add_command
class Version(Command):
    def add(self, parser):
        parser.add_parser("version", help="Display the kodman version information")

    def do(self, args, ctx, env, log):
        print(__version__)


def cli():
    engine.launch()


if __name__ == "__main__":
    cli()



import os
import subprocess
import sys
import tempfile
from pathlib import Path

import pytest

from data import responses

ENTRY_POINT = "kodman"
DOCKER_PROVIDER = os.getenv("DOCKER_PROVIDER", "podman")
KODMAN_SYSTEM_TESTING = os.getenv("KODMAN_SYSTEM_TESTING") == "true"


def remove_empty_lines(text):
    """Remove after solving https://github.com/epics-containers/Kodman/issues/9"""
    lines = [line for line in text.split("\n") if line.strip()]
    return "\n".join(lines)


@pytest.mark.skipif(
    not KODMAN_SYSTEM_TESTING, reason="export KODMAN_SYSTEM_TESTING=true"
)
def test_docker_run_hello():
    cmd = [DOCKER_PROVIDER, "run", "--rm", "hello-world"]
    assert subprocess.check_output(cmd).decode().strip() == responses.hello_world


@pytest.mark.skipif(
    not KODMAN_SYSTEM_TESTING, reason="export KODMAN_SYSTEM_TESTING=true"
)
def test_kodman_run_hello():
    cmd = [ENTRY_POINT, "run", "--rm", "hello-world"]
    assert subprocess.check_output(cmd).decode().strip() == remove_empty_lines(
        responses.hello_world
    )


@pytest.mark.skipif(
    not KODMAN_SYSTEM_TESTING, reason="export KODMAN_SYSTEM_TESTING=true"
)
def test_kodman_run_incluster(root: Path):
    pod_command = "pip install /kodman > /dev/null 2>&1 && kodman run --rm hello-world"
    cmd = [
        ENTRY_POINT,
        "run",
        "--rm",
        "--entrypoint",
        "bash",
        "-v",
        f"{root}:/kodman",
        f"python:{sys.version_info.major}.{sys.version_info.minor}",
        "-c",
        pod_command,
    ]

    result = subprocess.run(cmd, capture_output=True, text=True)
    assert result.stdout.strip() == remove_empty_lines(responses.hello_world)


@pytest.mark.skipif(
    not KODMAN_SYSTEM_TESTING, reason="export KODMAN_SYSTEM_TESTING=true"
)
def test_docker_run_exitcodes():
    ERROR_MSG = "My error message"
    cmd = [
        DOCKER_PROVIDER,
        "run",
        "--entrypoint",
        "bash",
        "--rm",
        "ubuntu",
        "-c",
        f"echo '{ERROR_MSG}' >&2; exit 1",
    ]

    result = subprocess.run(cmd, capture_output=True, text=True)
    assert result.returncode == 1
    assert result.stderr.strip() == ERROR_MSG


@pytest.mark.skipif(
    not KODMAN_SYSTEM_TESTING, reason="export KODMAN_SYSTEM_TESTING=true"
)
def test_kodman_run_exitcodes():
    ERROR_MSG = "My error message"
    cmd = [
        ENTRY_POINT,
        "run",
        "--entrypoint",
        "bash",
        "--rm",
        "ubuntu",
        "-c",
        f"echo '{ERROR_MSG}' >&2; exit 1",
    ]

    result = subprocess.run(cmd, capture_output=True, text=True)
    assert result.returncode == 1
    assert (
        result.stdout.strip() == ERROR_MSG
    )  # K8s does not distinguish between stderr and stdout!


@pytest.mark.skipif(
    not KODMAN_SYSTEM_TESTING, reason="export KODMAN_SYSTEM_TESTING=true"
)
def test_docker_run_mount_dir(data: Path):
    FILE_MOUNT = "to_mount.txt"
    cmd = [
        DOCKER_PROVIDER,
        "run",
        "-v",
        f"{data}:/test",
        "--rm",
        "ubuntu",
        "bash",
        "-c",
        f"cat test/{FILE_MOUNT}",
    ]
    assert subprocess.check_output(cmd).decode().strip() == responses.mount


@pytest.mark.skipif(
    not KODMAN_SYSTEM_TESTING, reason="export KODMAN_SYSTEM_TESTING=true"
)
def test_kodman_run_mount_dir(data: Path):
    FILE_MOUNT = "to_mount.txt"
    cmd = [
        ENTRY_POINT,
        "run",
        "-v",
        f"{data}:/test",
        "--rm",
        "ubuntu",
        "bash",
        "-c",
        f"cat test/{FILE_MOUNT}",
    ]

    assert subprocess.check_output(cmd).decode().strip() == responses.mount


def test_docker_run_mount_file(data: Path):
    FILE_MOUNT = "to_mount.txt"
    FILE_NEW = "to_read.txt"
    cmd = [
        DOCKER_PROVIDER,
        "run",
        "-v",
        f"{data}/{FILE_MOUNT}:/test/{FILE_NEW}",
        "--rm",
        "ubuntu",
        "bash",
        "-c",
        f"cat test/{FILE_NEW}",
    ]

    assert subprocess.check_output(cmd).decode().strip() == responses.mount


@pytest.mark.skipif(
    not KODMAN_SYSTEM_TESTING, reason="export KODMAN_SYSTEM_TESTING=true"
)
def test_kodman_run_mount_file(data: Path):
    FILE_MOUNT = "to_mount.txt"
    FILE_NEW = "to_read.txt"
    cmd = [
        ENTRY_POINT,
        "run",
        "-v",
        f"{data}/{FILE_MOUNT}:/test/{FILE_NEW}",
        "--rm",
        "ubuntu",
        "bash",
        "-c",
        f"cat test/{FILE_NEW}",
    ]

    assert subprocess.check_output(cmd).decode().strip() == responses.mount


@pytest.mark.skipif(
    not KODMAN_SYSTEM_TESTING, reason="export KODMAN_SYSTEM_TESTING=true"
)
def test_kodman_run_mount_root(data: Path):
    FILE_MOUNT = "to_mount.txt"
    FILE_NEW = "to_read.txt"
    cmd = [
        ENTRY_POINT,
        "run",
        "-v",
        f"{data}/{FILE_MOUNT}:/{FILE_NEW}",
        "--rm",
        "ubuntu",
        "bash",
        "-c",
        f"cat test/{FILE_NEW}",
    ]

    result = subprocess.run(cmd, capture_output=True, text=True)
    assert result.returncode == 1


@pytest.mark.skipif(
    not KODMAN_SYSTEM_TESTING, reason="export KODMAN_SYSTEM_TESTING=true"
)
def test_kodman_run_mount_large(data: Path):
    with tempfile.TemporaryDirectory() as tmpdirname:
        temp_dir = Path(tmpdirname)
        file_slug = "test.txt"
        file_name = temp_dir / file_slug
        with open(file_name, "wb") as out:
            out.truncate(100 * 1024 * 1024)

        cmd = [
            ENTRY_POINT,
            "run",
            "-v",
            f"{tmpdirname}:/test",
            "--rm",
            "ubuntu",
            "bash",
            "-c",
            f"[ -f test/{file_slug} ] && echo pass",
        ]

        assert subprocess.check_output(cmd).decode().strip() == "pass"


@pytest.mark.skipif(
    not KODMAN_SYSTEM_TESTING, reason="export KODMAN_SYSTEM_TESTING=true"
)
def test_kodman_fail_image(data: Path):
    cmd = [
        ENTRY_POINT,
        "run",
        "--rm",
        "hello-worl",
    ]

    result = subprocess.run(cmd, capture_output=True, text=True)
    assert result.returncode == 1
    assert responses.failed_image in result.stderr


@pytest.mark.skipif(
    not KODMAN_SYSTEM_TESTING, reason="export KODMAN_SYSTEM_TESTING=true"
)
def test_kodman_fail_command(data: Path):
    cmd = [
        ENTRY_POINT,
        "run",
        "--rm",
        "--entrypoint",
        "bash",
        "hello-world",
    ]

    result = subprocess.run(cmd, capture_output=True, text=True)
    assert result.returncode == 1
    assert responses.failed_command in result.stderr



import os
from typing import overload


@overload
def get_env(variable: str, expected_type: type[bool]) -> bool | None: ...


@overload
def get_env(variable: str, expected_type: type[int]) -> int | None: ...


@overload
def get_env(variable: str, expected_type: type[str]) -> str | None: ...


def get_env(
    variable: str, expected_type: type[bool | int | str]
) -> bool | int | str | None:
    val = os.getenv(variable)

    if val is None:
        val = None
    elif expected_type is bool:
        if val.lower() in ("true", "1"):
            val = True
        elif val.lower() in ("false", "0"):
            val = False
        else:
            val = None
    elif expected_type is int:
        try:
            val = int(val)
        except ValueError:
            val = None
    elif expected_type is str:
        pass
    else:
        raise TypeError(f"Unsupported type: {expected_type}")

    return val



import os
from pathlib import Path
from typing import Any

import pytest

# Prevent pytest from catching exceptions when debugging in vscode so that break on
# exception works correctly (see: https://github.com/pytest-dev/pytest/issues/7409)
if os.getenv("PYTEST_RAISE", "0") == "1":

    @pytest.hookimpl(tryfirst=True)
    def pytest_exception_interact(call: pytest.CallInfo[Any]):
        if call.excinfo is not None:
            raise call.excinfo.value
        else:
            raise RuntimeError(
                f"{call} has no exception data, an unknown error has occurred"
            )

    @pytest.hookimpl(tryfirst=True)
    def pytest_internalerror(excinfo: pytest.ExceptionInfo[Any]):
        raise excinfo.value


DATA_PATH = Path(__file__).parent / "data"


@pytest.fixture
def data() -> Path:
    return DATA_PATH


ROOT_PATH = Path(__file__).parent.parent


@pytest.fixture
def root() -> Path:
    return ROOT_PATH


@pytest.fixture
def env_vars(mocker):
    mocker.patch.dict(
        os.environ,
        {
            "KODMAN_TEST_STRING": "Test",
            "KODMAN_TEST_BOOL_TRUE": "true",
            "KODMAN_TEST_BOOL_FALSE": "false",
            "KODMAN_TEST_INT": "99",
        },
    )



help_screen = """usage: kodman [-h] [-v] {run,version} ...

positional arguments:
  {run,version}
    run          Run a command in a new container
    version      Display the kodman version information

options:
  -h, --help     show this help message and exit
  -v, --version  show program's version number and exit

environment variables:
  KODMAN_DEBUG  bool
  KODMAN_SERVICE_ACCOUNT  str"""

hello_world = """Hello from Docker!
This message shows that your installation appears to be working correctly.

To generate this message, Docker took the following steps:
 1. The Docker client contacted the Docker daemon.
 2. The Docker daemon pulled the "hello-world" image from the Docker Hub.
    (amd64)
 3. The Docker daemon created a new container from that image which runs the
    executable that produces the output you are currently reading.
 4. The Docker daemon streamed that output to the Docker client, which sent it
    to your terminal.

To try something more ambitious, you can run an Ubuntu container with:
 $ docker run -it ubuntu bash

Share images, automate workflows, and more with a free Docker ID:
 https://hub.docker.com/

For more examples and ideas, visit:
 https://docs.docker.com/get-started/"""

mount = "test data"

failed_image = "Failed to pull image"

failed_command = 'exec: "bash": executable file not found in $PATH: unknown'



import argparse
import logging
import sys
from abc import ABC, abstractmethod
from typing import overload

from rich.console import Console

from .utilities import get_env as _get_env


class Command(ABC):
    exit_code = 0

    @abstractmethod
    def do(self, args, ctx, env, log):
        pass

    @abstractmethod
    def add(self, parser):
        pass


class ConsoleOutputHandler(logging.Handler):
    def __init__(self, status):
        super().__init__()
        self._status = status

    def emit(self, record):
        self._status.update(record.msg)


class ArgparseEngine:
    _env_types = {}
    _env_vals = {}

    def __init__(self, debug=False):
        # Configure logging
        self._log = logging.getLogger("ArgparseEngine")
        self._console = Console()
        self._status = None
        if debug:
            formatter = logging.Formatter("%(levelname)s:\t%(message)s")
            handler = logging.StreamHandler()
            handler.setFormatter(formatter)
            self._log.addHandler(handler)
            self._log.setLevel("DEBUG")
        else:
            self._status = self._console.status("Initializing application...")
            handler = ConsoleOutputHandler(self._status)
            self._log.addHandler(handler)
            self._log.setLevel("INFO")

        # Configure application
        self._parser = argparse.ArgumentParser(
            formatter_class=argparse.RawTextHelpFormatter,
        )
        self._subparsers = self._parser.add_subparsers(dest="cli_command")
        self._ctx = None
        self._args = []
        self._commands = []

    def add_command(self, command: type[Command]):
        self._commands.append(command())

    @overload
    def get_env(self, variable: str, expected_type: type[bool]) -> bool | None: ...

    @overload
    def get_env(self, variable: str, expected_type: type[int]) -> int | None: ...

    @overload
    def get_env(self, variable: str, expected_type: type[str]) -> str | None: ...

    def get_env(
        self, variable: str, expected_type: type[bool | int | str]
    ) -> bool | int | str | None:
        val = _get_env(variable, expected_type)
        self._env_types[variable] = expected_type
        self._env_vals[variable] = val

        return val

    def _process_env(self):
        message = "environment variables:\n"
        for env, _type in self._env_types.items():
            message_prefix = f"  {env}  {_type.__name__}"
            message_body = (
                f"  (current: {self._env_vals[env]})" if self._env_vals[env] else ""
            )
            message_suffix = "\n"
            message += message_prefix + message_body + message_suffix
        self._parser.epilog = message

    def launch(self):
        for command in self._commands:
            command.add(self._subparsers)

        self._process_env()

        args = self._parser.parse_args()

        if self._status:
            self._status.start()

        for command in self._commands:
            if args.cli_command == command.__class__.__name__.lower():
                command.do(args, self._ctx, self._env_vals, self._log)

                if self._status:
                    self._status.stop()

                sys.exit(command.exit_code)



import subprocess

from data import responses
from kodman import __version__

ENTRY_POINT = "kodman"


def test_cli_version():
    cmd = [ENTRY_POINT, "--version"]
    assert subprocess.check_output(cmd).decode().strip() == __version__


def test_cli_help():
    cmd = [ENTRY_POINT, "--help"]
    assert subprocess.check_output(cmd).decode().strip() == responses.help_screen



import io
import logging
import sys
import tarfile
import time
from dataclasses import dataclass, field
from pathlib import Path
from typing import Any

from kubernetes import client, config, watch
from kubernetes.client.models.v1_pod import V1Pod
from kubernetes.client.rest import ApiException
from kubernetes.stream import stream


@dataclass(frozen=True)
class RunOptions:
    image: str
    command: list[str] = field(default_factory=lambda: [])
    args: list[str] = field(default_factory=lambda: [])
    volumes: list[str] = field(default_factory=lambda: [])
    service_account: str = field(default_factory=lambda: "")

    def __hash__(self):
        hash_candidates = (
            self.image,
            self.command,
            self.args,
            self.volumes,
            time.time(),  # Add timestamp
        )

        to_hash = []
        for item in hash_candidates:
            if not item:  # Skip unhashable falsy items
                pass
            elif type(item) is list:  # Make hashable
                to_hash.append(tuple(item))
            else:
                to_hash.append(item)

        _hash = hash(tuple(to_hash))
        _hash += sys.maxsize + 1  # Ensure always positive
        return _hash


@dataclass(frozen=True)
class DeleteOptions:
    name: str


def cp_k8s(
    kube_conn: client.CoreV1Api,
    namespace: str,
    pod_name: str,
    container: str,
    source_path: Path,
    dest_path: Path,
    log: logging.Logger,
):
    log.info(f"Transferring {source_path} to {dest_path}")
    buf = io.BytesIO()

    log.debug(f"Compressing {source_path}")
    with tarfile.open(fileobj=buf, mode="w:tar") as tar:  # To compress set 'w:gz'
        tar.add(source_path, arcname=dest_path)
    buf.seek(0)
    compressed_size = buf.getbuffer().nbytes
    log.debug(f"Compressed to {compressed_size} bytes")

    exec_command = ["tar", "xvf", "-", "-C", "/"]  # To decompress set 'xzvf'
    resp = stream(
        kube_conn.connect_get_namespaced_pod_exec,
        pod_name,
        namespace,
        container=container,
        command=exec_command,
        stderr=True,
        stdin=True,
        stdout=True,
        tty=False,
        _preload_content=False,
    )

    chunk_size = 10 * 1024 * 1024
    steps = -(compressed_size // -chunk_size)  # Ceiling division
    log.debug(f"Transferring {steps} chunks")
    counter = 0
    while resp.is_open():
        resp.update(timeout=1)
        if read := buf.read(chunk_size):
            resp.write_stdin(read)
        else:
            log.debug("Empty buffer")
            break
        log.info(f"Transfer {counter * 100 // steps}% completed")
        counter += 1
    resp.close()
    log.info("Transfer done")


def get_incluster_context():
    ns_path = "/var/run/secrets/kubernetes.io/serviceaccount/namespace"
    context = {}
    with open(ns_path) as f:
        context["namespace"] = f.read().strip()
    context["cluster"] = "default"
    context["user"] = "default"
    return context


class Backend:
    def __init__(self, log):
        self.return_code = 0
        self._log = log
        self._polling_freq = 1
        self._grace_period = 2  # Is this too aggressive?

    def connect(self):
        # Load config for user/serviceaccount
        # https://github.com/kubernetes-client/python/issues/1005
        try:
            self._log.info(
                "Loading kube config for user interaction from outside of cluster"
            )
            config.load_kube_config()
            self._log.info("Loaded kube config successfully")
            self._context = config.list_kube_config_contexts()[1]["context"]
        except config.config_exception.ConfigException:
            self._log.info("Failed to load kube config, trying in-cluster config")
            config.load_incluster_config()
            self._log.info("Loaded in-cluster config successfully")
            self._context = get_incluster_context()

        self._client = client.CoreV1Api()
        self._log.debug("The current context is:")
        self._log.debug(f"  Cluster: {self._context['cluster']}")
        self._log.debug(f"  Namespace: {self._context['namespace']}")
        self._log.debug(f"  User: {self._context['user']}")

    def run(self, options: RunOptions) -> str:
        unique_pod_name = f"kodman-run-{hash(options)}"
        init_container_name = "wait-for-signal"
        namespace = self._context["namespace"]
        pod_manifest: dict[str, Any] = {
            "apiVersion": "v1",
            "kind": "Pod",
            "metadata": {
                "name": unique_pod_name,
            },
            "spec": {
                "initContainers": [
                    {
                        "name": init_container_name,
                        "image": "busybox",
                        "command": [
                            "sh",
                            "-c",
                            "until [ -f /tmp/trigger ];"
                            'do echo "Waiting for trigger...";'
                            "sleep 1;"
                            "done;"
                            'echo "Trigger file found!"',
                        ],
                        "volumeMounts": [],
                    },
                ],
                "containers": [
                    {
                        "image": options.image,
                        "name": "kodman-exec",
                        "volumeMounts": [],
                    }
                ],
                "volumes": [],
            },
        }

        if options.command:
            container = pod_manifest["spec"]["containers"][0]
            container["command"] = options.command

        if options.args:
            pod_manifest["spec"]["containers"][0]["args"] = options.args

        if options.service_account:
            self._log.debug(f"Using serviceAccountNam: '{options.service_account}'")
            pod_manifest["spec"]["serviceAccountName"] = options.service_account

        volumes: list[dict[str, Path]] = []
        if options.volumes:
            for i, options_volume in enumerate(options.volumes):
                process = options_volume.split(":")
                src = Path(process[0]).resolve()
                if not src.exists():
                    raise FileNotFoundError(f"{src} does not exist")
                dst = src  # In case no dst, set same as src
                try:
                    dst = Path(process[1])
                except IndexError:
                    pass
                if not dst.is_absolute():
                    raise ValueError("Destination path must be absolute")
                self._log.info(f"Mount: {src} to {dst}")
                if src.is_dir():
                    self._log.debug(f"Volume target {src} is a directory")
                    dst_mount = dst
                else:
                    self._log.debug(f"Volume target {src} is a file")
                    dst_mount = dst.parent
                    if dst_mount == Path("/"):
                        raise NotImplementedError(
                            "Root mounting of files not supported by k8s 'emptyDir'"
                        )

                volumes.append({"src": src, "dst": dst})  # cache for later

                pod_manifest["spec"]["initContainers"][0]["volumeMounts"].append(
                    {"name": f"shared-data-{i}", "mountPath": str(dst_mount)}
                )
                pod_manifest["spec"]["containers"][0]["volumeMounts"].append(
                    {"name": f"shared-data-{i}", "mountPath": str(dst_mount)}
                )
                pod_manifest["spec"]["volumes"].append(
                    {
                        "name": f"shared-data-{i}",
                        "emptyDir": {},
                    }
                )

        self._log.debug(f"Pod manifest = {pod_manifest}")

        # Schedule pod and block until ready
        self._log.info(f"Creating pod: {unique_pod_name}")
        self._client.create_namespaced_pod(body=pod_manifest, namespace=namespace)
        while True:
            read_resp = self._client.read_namespaced_pod(
                name=unique_pod_name, namespace=namespace
            )
            # Runtime type checking
            if isinstance(read_resp, V1Pod):
                if not read_resp.status:
                    raise ValueError("Empty pod status")
                if read_resp.status.init_container_statuses:
                    init_status = read_resp.status.init_container_statuses[0]
                    if init_status.state.running:
                        self._log.info("Init container is running")
                        break
            else:
                raise TypeError("Unexpected response type")

            self._log.info("Awaiting init container...")
            time.sleep(1 / self._polling_freq)

        # Fill volumes
        for volume in volumes:
            cp_k8s(
                self._client,
                namespace,
                unique_pod_name,
                init_container_name,
                volume["src"],
                volume["dst"],
                log=self._log,
            )

        # Start execution
        self._log.info("Execution start")
        exec_command = [
            "/bin/sh",
            "-c",
            "touch /tmp/trigger",
        ]
        _ = stream(
            self._client.connect_get_namespaced_pod_exec,
            unique_pod_name,
            namespace,
            container=init_container_name,
            command=exec_command,
            stderr=True,
            stdin=False,
            stdout=True,
            tty=False,
        )

        while True:
            read_resp = self._client.read_namespaced_pod(
                name=unique_pod_name, namespace=namespace
            )
            if isinstance(read_resp, V1Pod):  # Runtime type checking
                if not read_resp.status:
                    raise ValueError("Empty pod status")
                elif read_resp.status.phase != "Pending":
                    self._log.info(f"Pod status: {read_resp.status.phase}")
                    break
                self._log.info(f"Pod status: {read_resp.status.phase}")
                time.sleep(1 / self._polling_freq)
                events = self._client.list_namespaced_event(
                    namespace=namespace,
                    field_selector=f"involvedObject.name={unique_pod_name}",
                )
                for event in events.items:
                    if event.type == "Warning":
                        self.return_code = 1
                        reason = event.type
                        message = event.message
                        self._log.debug(f"{reason}: {message}")
                        print(message, file=sys.stderr)
                        return unique_pod_name
            else:
                raise TypeError("Unexpected response type")

        # Attach to pod logging
        self._log.info("Try attach to pod logs")
        w = watch.Watch()
        for e in w.stream(
            self._client.read_namespaced_pod_log,
            name=unique_pod_name,
            namespace=namespace,
            follow=True,
        ):
            print(e)
        self._log.info("Execution complete")
        w.stop()

        # Check exit codes
        final_pod = self._client.read_namespaced_pod(
            name=unique_pod_name,
            namespace=namespace,
        )
        if isinstance(final_pod, V1Pod):  # Runtime type checking
            if not final_pod.status:
                raise ValueError("Empty pod status")
            container_status = final_pod.status.container_statuses[0]
            while not container_status.state.terminated:
                # Exit early if container didnt even start
                if not container_status.started:
                    self._log.info("Container failed to start")
                    self.return_code = 1
                    reason = container_status.state.waiting.reason
                    message = container_status.state.waiting.message
                    self._log.debug(f"{reason}: {message}")
                    print(message, file=sys.stderr)
                    return unique_pod_name

                self._log.info("Awaiting pod termination...")
                time.sleep(1 / self._polling_freq)
                final_pod = self._client.read_namespaced_pod(
                    name=unique_pod_name,
                    namespace=namespace,
                )
                container_status = final_pod.status.container_statuses[0]  # type: ignore
            self.return_code = container_status.state.terminated.exit_code

        return unique_pod_name

    def delete(self, options: DeleteOptions):
        namespace = self._context["namespace"]
        try:
            exists_resp = self._client.read_namespaced_pod(
                name=options.name,
                namespace=namespace,
            )
            self._client.delete_namespaced_pod(
                name=options.name,
                namespace=namespace,
                grace_period_seconds=self._grace_period,
            )
            while exists_resp:
                self._log.info("Awaiting pod cleanup...")
                try:
                    exists_resp = self._client.read_namespaced_pod(
                        name=options.name,
                        namespace=namespace,
                    )
                    time.sleep(1 / self._polling_freq)
                except ApiException as e:
                    if e.status == 404:
                        self._log.info(f"Pod {options.name} deleted successfully")
                        break
                    else:
                        raise e

        except ApiException as e:
            self._log.info(f"Error deleting pod: {e}")



from kodman.utilities import get_env


def test_get_env_string(env_vars):
    var = get_env("KODMAN_TEST_STRING", str)
    assert var == "Test"


def test_get_env_bool_true(env_vars):
    var = get_env("KODMAN_TEST_BOOL_TRUE", bool)
    assert var is True


def test_get_env_bool_false(env_vars):
    var = get_env("KODMAN_TEST_BOOL_FALSE", bool)
    assert var is False


def test_get_env_int(env_vars):
    var = get_env("KODMAN_TEST_INT", int)
    assert var == 99


def test_get_env_none(env_vars):
    var = get_env("KODMAN_TEST_NONE", int)
    assert var is None
